@@ -37,7 +37,7 @@
 			   (p)->type == para_Title ? -1 : 0 )
 
 typedef struct {
-    int just_numbers;
+    int number_at_all, just_numbers;
     wchar_t *number_suffix;
 } sectlevel;
 
@@ -61,7 +61,7 @@ typedef struct {
     char *body_tag, *nav_attr;
     wchar_t *author, *description;
     wchar_t *index_text, *contents_text, *preamble_text, *title_separator;
-    wchar_t *nav_prev_text, *nav_next_text, *nav_separator;
+    wchar_t *nav_prev_text, *nav_next_text, *nav_up_text, *nav_separator;
     wchar_t *index_main_sep, *index_multi_sep;
     wchar_t *pre_versionid, *post_versionid;
     int restrict_charset, output_charset;
@@ -248,10 +248,12 @@ static htmlconfig html_configure(paragraph *source) {
      */
     ret.leaf_level = 2;
     ret.achapter.just_numbers = FALSE;
+    ret.achapter.number_at_all = TRUE;
     ret.achapter.number_suffix = L": ";
     ret.nasect = 1;
     ret.asect = snewn(ret.nasect, sectlevel);
     ret.asect[0].just_numbers = TRUE;
+    ret.asect[0].number_at_all = TRUE;
     ret.asect[0].number_suffix = L" ";
     ret.ncdepths = 0;
     ret.contents_depths = 0;
@@ -282,6 +284,7 @@ static htmlconfig html_configure(paragraph *source) {
     ret.title_separator = L" - ";
     ret.nav_prev_text = L"Previous";
     ret.nav_next_text = L"Next";
+    ret.nav_up_text = L"Up";
     ret.nav_separator = L" | ";
     ret.index_main_sep = L": ";
     ret.index_multi_sep = L", ";
@@ -345,7 +348,7 @@ static htmlconfig html_configure(paragraph *source) {
 			break;
 
 		if (i == lenof(versions))
-		    error(err_htmlver, &p->fpos, vername);
+		    err_htmlver(&p->fpos, vername);
 		else
 		    ret.htmlver = versions[i].ver;
 	    } else if (!ustricmp(k, L"html-single-filename")) {
@@ -378,9 +381,11 @@ static htmlconfig html_configure(paragraph *source) {
 			frag = adv(frag);
 		    }
 		} else
-		    error(err_cfginsufarg, &p->fpos, p->origkeyword, 1);
+		    err_cfginsufarg(&p->fpos, p->origkeyword, 1);
 	    } else if (!ustricmp(k, L"html-chapter-numeric")) {
 		ret.achapter.just_numbers = utob(uadv(k));
+	    } else if (!ustricmp(k, L"html-chapter-shownumber")) {
+		ret.achapter.number_at_all = utob(uadv(k));
 	    } else if (!ustricmp(k, L"html-suppress-navlinks")) {
 		ret.navlinks = !utob(uadv(k));
 	    } else if (!ustricmp(k, L"html-rellinks")) {
@@ -410,6 +415,21 @@ static htmlconfig html_configure(paragraph *source) {
 		    ret.nasect = n+1;
 		}
 		ret.asect[n].just_numbers = utob(q);
+	    } else if (!ustricmp(k, L"html-section-shownumber")) {
+		wchar_t *q = uadv(k);
+		int n = 0;
+		if (uisdigit(*q)) {
+		    n = utoi(q);
+		    q = uadv(q);
+		}
+		if (n >= ret.nasect) {
+		    int i;
+		    ret.asect = sresize(ret.asect, n+1, sectlevel);
+		    for (i = ret.nasect; i <= n; i++)
+			ret.asect[i] = ret.asect[ret.nasect-1];
+		    ret.nasect = n+1;
+		}
+		ret.asect[n].number_at_all = utob(q);
 	    } else if (!ustricmp(k, L"html-section-suffix")) {
 		wchar_t *q = uadv(k);
 		int n = 0;
@@ -494,6 +514,8 @@ static htmlconfig html_configure(paragraph *source) {
 		ret.nav_prev_text = uadv(k);
 	    } else if (!ustricmp(k, L"html-nav-next-text")) {
 		ret.nav_next_text = uadv(k);
+	    } else if (!ustricmp(k, L"html-nav-up-text")) {
+		ret.nav_up_text = uadv(k);
 	    } else if (!ustricmp(k, L"html-nav-separator")) {
 		ret.nav_separator = uadv(k);
 	    } else if (!ustricmp(k, L"html-index-main-separator")) {
@@ -526,7 +548,7 @@ static htmlconfig html_configure(paragraph *source) {
      * turn both off.
      */
     if (!ret.chm_filename ^ !ret.hhp_filename) {
-	error(err_chmnames);
+	err_chmnames();
 	sfree(ret.chm_filename); ret.chm_filename = NULL;
 	sfree(ret.hhp_filename); ret.hhp_filename = NULL;
     }
@@ -862,9 +884,12 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 #define listname(lt) ( (lt)==UL ? "ul" : (lt)==OL ? "ol" : "dl" )
 #define itemname(lt) ( (lt)==LI ? "li" : (lt)==DT ? "dt" : "dd" )
 
-	    ho.fp = fopen(f->filename, "w");
+	    if (!strcmp(f->filename, "-"))
+		ho.fp = stdout;
+	    else
+		ho.fp = fopen(f->filename, "w");
 	    if (!ho.fp)
-		error(err_cantopenw, f->filename);
+		err_cantopenw(f->filename);
 
 	    ho.charset = conf.output_charset;
 	    ho.restrict_charset = conf.restrict_charset;
@@ -970,8 +995,6 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 		    html_nl(&ho);
 		}
 
-		/* FIXME: link rel="up" */
-
 		if (f != files.head) {
 		    element_empty(&ho, "link");
 		    element_attr(&ho, "rel", "ToC");
@@ -979,6 +1002,17 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 		    html_nl(&ho);
 		}
 
+		if (conf.leaf_level > 0) {
+		    htmlsect *p = f->first->parent;
+		    assert(p == f->last->parent);
+		    if (p) {
+			element_empty(&ho, "link");
+			element_attr(&ho, "rel", "up");
+			element_attr(&ho, "href", p->file->filename);
+			html_nl(&ho);
+		    }
+		}
+
 		if (has_index && files.index && f != files.index) {
 		    element_empty(&ho, "link");
 		    element_attr(&ho, "rel", "index");
@@ -1057,6 +1091,22 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 		if (f != files.head)
 		    element_close(&ho, "a");
 
+		/* We don't bother with "Up" links for leaf-level 1,
+		 * as they would be identical to the "Contents" links. */
+		if (conf.leaf_level >= 2) {
+		    htmlsect *p = f->first->parent;
+		    assert(p == f->last->parent);
+		    html_text(&ho, conf.nav_separator);
+		    if (p) {
+			element_open(&ho, "a");
+			element_attr(&ho, "href", p->file->filename);
+		    }
+		    html_text(&ho, conf.nav_up_text);
+		    if (p) {
+			element_close(&ho, "a");
+		    }
+		}
+
 		if (has_index && files.index) {
 		    html_text(&ho, conf.nav_separator);
 		    if (f != files.index) {
@@ -1083,6 +1133,28 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 	    }
 	    prevf = f;
 
+	    /*
+	     * Special case: for single-file mode, we output the top
+	     * section title before the TOC.
+	     */
+	    if (files.head == files.tail && sects.head->type == TOP) {
+		element_open(&ho, "h1");
+
+		/*
+		 * Provide anchor(s) for cross-links to target.
+		 */
+		{
+		    int i;
+		    for (i=0; i < conf.ntfragments; i++)
+			if (sects.head->fragments[i])
+			    html_fragment(&ho, sects.head->fragments[i]);
+		}
+
+		html_section_title(&ho, sects.head, f, keywords, &conf, TRUE);
+
+		element_close(&ho, "h1");
+	    }
+
 	    /*
 	     * Write out a prefix TOC for the file (if a leaf file).
 	     * 
@@ -1093,12 +1165,9 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 	     * contains all descendants of any section it
 	     * contains), because this will play a part in our
 	     * decision on whether or not to _output_ the TOC.
-	     * 
-	     * Special case: we absolutely do not do this if we're
-	     * in single-file mode.
 	     */
-	    if (files.head != files.tail) {
-		int ntoc = 0, tocsize = 0;
+	    {
+		int ntoc = 0, tocsize = 0, tocstartidx = 0;
 		htmlsect **toc = NULL;
 		int leaf = TRUE;
 
@@ -1135,16 +1204,28 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 		    }
 		}
 
+		/*
+		 * Special case: for single-file mode, we don't output
+		 * the first section TOC entry if it's the top since we
+		 * have already output a section title for it above the
+		 * TOC, and since we don't output the top TOC entry, we
+		 * reduce the level of the remaining TOC entries by one
+		 * so that they are output correctly one level up from
+		 * where they would have been.
+		 */
+		tocstartidx = (files.head == files.tail && ntoc > 0 &&
+			       toc[0]->type == TOP) ? 1 : 0;
 		if (leaf && conf.leaf_contains_contents &&
-		    ntoc >= conf.leaf_smallest_contents) {
+		    ntoc >= conf.leaf_smallest_contents &&
+		    tocstartidx < ntoc) {
 		    int i;
 
-		    for (i = 0; i < ntoc; i++) {
+		    for (i = tocstartidx; i < ntoc; i++) {
 			htmlsect *s = toc[i];
 			int hlevel = (s->type == TOP ? -1 :
 				      s->type == INDEX ? 0 :
 				      heading_depth(s->title))
-			    - f->min_heading_depth + 1;
+			    - tocstartidx - f->min_heading_depth + 1;
 
 			assert(hlevel >= 1);
 			html_contents_entry(&ho, hlevel, s,
@@ -1216,39 +1297,44 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 
 		    /*
 		     * Display the section heading.
+		     *
+		     * Special case: for single-file mode, we don't
+		     * output the top section title since we have
+		     * already output it before the TOC.
 		     */
+		    if (files.head != files.tail || s->type != TOP) {
+			hlevel = (s->type == TOP ? -1 :
+				  s->type == INDEX ? 0 :
+				  heading_depth(s->title))
+			    - f->min_heading_depth + 1;
+			assert(hlevel >= 1);
+			/* HTML headings only go up to <h6> */
+			if (hlevel > 6)
+			    hlevel = 6;
+			htag[0] = 'h';
+			htag[1] = '0' + hlevel;
+			htag[2] = '\0';
+			element_open(&ho, htag);
 
-		    hlevel = (s->type == TOP ? -1 :
-			      s->type == INDEX ? 0 :
-			      heading_depth(s->title))
-			- f->min_heading_depth + 1;
-		    assert(hlevel >= 1);
-		    /* HTML headings only go up to <h6> */
-		    if (hlevel > 6)
-			hlevel = 6;
-		    htag[0] = 'h';
-		    htag[1] = '0' + hlevel;
-		    htag[2] = '\0';
-		    element_open(&ho, htag);
+			/*
+			 * Provide anchor(s) for cross-links to target.
+			 * 
+			 * (Also we'll have to do this separately in
+			 * other paragraph types - NumberedList and
+			 * BiblioCited.)
+			 */
+			{
+			    int i;
+			    for (i=0; i < conf.ntfragments; i++)
+				if (s->fragments[i])
+				    html_fragment(&ho, s->fragments[i]);
+			}
 
-		    /*
-		     * Provide anchor(s) for cross-links to target.
-		     * 
-		     * (Also we'll have to do this separately in
-		     * other paragraph types - NumberedList and
-		     * BiblioCited.)
-		     */
-		    {
-			int i;
-			for (i=0; i < conf.ntfragments; i++)
-			    if (s->fragments[i])
-				html_fragment(&ho, s->fragments[i]);
+			html_section_title(&ho, s, f, keywords, &conf, TRUE);
+
+			element_close(&ho, htag);
 		    }
 
-		    html_section_title(&ho, s, f, keywords, &conf, TRUE);
-
-		    element_close(&ho, htag);
-
 		    /*
 		     * Now display the section text.
 		     */
@@ -1635,7 +1721,7 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 
 	ho.fp = fopen(conf.hhp_filename, "w");
 	if (!ho.fp)
-	    error(err_cantopenw, conf.hhp_filename);
+	    err_cantopenw(conf.hhp_filename);
 
 	fprintf(ho.fp,
 		"[OPTIONS]\n"
@@ -1710,7 +1796,7 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 
 	ho.fp = fopen(conf.hhc_filename, "w");
 	if (!ho.fp)
-	    error(err_cantopenw, conf.hhc_filename);
+	    err_cantopenw(conf.hhc_filename);
 
 	ho.charset = CS_CP1252;	       /* as far as I know, HHC files are */
 	ho.restrict_charset = CS_CP1252;   /* hardwired to this charset */
@@ -1824,7 +1910,7 @@ void html_backend(paragraph *sourceform, keywordlist *keywords,
 
 	ho.fp = fopen(hhk_filename, "w");
 	if (!ho.fp)
-	    error(err_cantopenw, hhk_filename);
+	    err_cantopenw(hhk_filename);
 
 	ho.charset = CS_CP1252;	       /* as far as I know, HHK files are */
 	ho.restrict_charset = CS_CP1252;   /* hardwired to this charset */
@@ -2126,7 +2212,7 @@ static void html_words(htmloutput *ho, word *words, int flags,
 		       htmlfile *file, keywordlist *keywords, htmlconfig *cfg)
 {
     word *w;
-    char *c;
+    char *c, *c2, *p, *q;
     int style, type;
 
     for (w = words; w; w = w->next) switch (w->type) {
@@ -2134,7 +2220,20 @@ static void html_words(htmloutput *ho, word *words, int flags,
 	if (flags & LINKS) {
 	    element_open(ho, "a");
 	    c = utoa_dup(w->text, CS_ASCII);
-	    element_attr(ho, "href", c);
+	    c2 = snewn(1 + 10*strlen(c), char);
+	    for (p = c, q = c2; *p; p++) {
+		if (*p == '&')
+		    q += sprintf(q, "&amp;");
+		else if (*p == '<')
+		    q += sprintf(q, "&lt;");
+		else if (*p == '>')
+		    q += sprintf(q, "&gt;");
+		else
+		    *q++ = *p;
+	    }
+	    *q = '\0';
+	    element_attr(ho, "href", c2);
+	    sfree(c2);
 	    sfree(c);
 	}
 	break;
@@ -2145,13 +2244,23 @@ static void html_words(htmloutput *ho, word *words, int flags,
 	    paragraph *p;
 	    htmlsect *s;
 
-	    assert(kwl);
-	    p = kwl->para;
-	    s = (htmlsect *)p->private_data;
+            /* kwl should be non-NULL in any sensible case, but if the
+             * input contained an unresolvable xref then it might be
+             * NULL as an after-effect of that */
+	    if (kwl) {
+                p = kwl->para;
+                s = (htmlsect *)p->private_data;
 
-	    assert(s);
+                assert(s);
 
-	    html_href(ho, file, s->file, s->fragments[0]);
+                html_href(ho, file, s->file, s->fragments[0]);
+            } else {
+                /* If kwl is NULL, we must still open an href to
+                 * _somewhere_, because it's easier than remembering
+                 * which one not to close when we come to the
+                 * XrefEnd */
+                html_href(ho, file, file, NULL);
+            }
 	}
 	break;
       case word_HyperEnd:
@@ -2168,14 +2277,17 @@ static void html_words(htmloutput *ho, word *words, int flags,
 	break;
       case word_Normal:
       case word_Emph:
+      case word_Strong:
       case word_Code:
       case word_WeakCode:
       case word_WhiteSpace:
       case word_EmphSpace:
+      case word_StrongSpace:
       case word_CodeSpace:
       case word_WkCodeSpace:
       case word_Quote:
       case word_EmphQuote:
+      case word_StrongQuote:
       case word_CodeQuote:
       case word_WkCodeQuote:
 	style = towordstyle(w->type);
@@ -2185,6 +2297,11 @@ static void html_words(htmloutput *ho, word *words, int flags,
 	     attraux(w->aux) == attr_Only) &&
 	    (flags & MARKUP))
 	    element_open(ho, "em");
+	else if (style == word_Strong &&
+	    (attraux(w->aux) == attr_First ||
+	     attraux(w->aux) == attr_Only) &&
+	    (flags & MARKUP))
+	    element_open(ho, "strong");
 	else if ((style == word_Code || style == word_WeakCode) &&
 		 (attraux(w->aux) == attr_First ||
 		  attraux(w->aux) == attr_Only) &&
@@ -2210,6 +2327,11 @@ static void html_words(htmloutput *ho, word *words, int flags,
 	     attraux(w->aux) == attr_Only) &&
 	    (flags & MARKUP))
 	    element_close(ho, "em");
+	else if (style == word_Strong &&
+	    (attraux(w->aux) == attr_Last ||
+	     attraux(w->aux) == attr_Only) &&
+	    (flags & MARKUP))
+	    element_close(ho, "strong");
 	else if ((style == word_Code || style == word_WeakCode) &&
 		 (attraux(w->aux) == attr_Last ||
 		  attraux(w->aux) == attr_Only) &&
@@ -2460,7 +2582,7 @@ static void html_text_limit_internal(htmloutput *ho, wchar_t const *text,
 static void cleanup(htmloutput *ho)
 {
     return_to_neutral(ho);
-    if (ho->fp)
+    if (ho->fp && ho->fp != stdout)
 	fclose(ho->fp);
 }
 
@@ -2738,7 +2860,7 @@ static void html_section_title(htmloutput *ho, htmlsect *s, htmlfile *thisfile,
 	else
 	    sl = &cfg->asect[cfg->nasect-1];
 
-	if (!sl)
+	if (!sl || !sl->number_at_all)
 	    number = NULL;
 	else if (sl->just_numbers)
 	    number = s->title->kwtext2;
