@@ -215,8 +215,9 @@ paragraph *man_config_filename(char *filename)
     return cmdline_cfg_simple("man-filename", filename, NULL);
 }
 
-#define QUOTE_INITCTRL 1 /* quote initial . and ' on a line */
-#define QUOTE_QUOTES   2 /* quote double quotes by doubling them */
+#define QUOTE_INITCTRL    1 /* quote initial . and ' on a line */
+#define QUOTE_QUOTES      2 /* quote double quotes by doubling them */
+#define QUOTE_LITERAL     4 /* defeat special meaning of `, ', - in troff */
 
 void man_backend(paragraph *sourceform, keywordlist *keywords,
 		 indexdata *idx, void *unused) {
@@ -234,9 +235,12 @@ void man_backend(paragraph *sourceform, keywordlist *keywords,
     /*
      * Open the output file.
      */
-    fp = fopen(conf.filename, "w");
+    if (!strcmp(conf.filename, "-"))
+	fp = stdout;
+    else
+	fp = fopen(conf.filename, "w");
     if (!fp) {
-	error(err_cantopenw, conf.filename);
+	err_cantopenw(conf.filename);
 	return;
     }
 
@@ -247,6 +251,12 @@ void man_backend(paragraph *sourceform, keywordlist *keywords,
 	    man_text(fp, p->words, TRUE, 0, &conf);
 	}
 
+    /* Standard preamble */
+    /* Dodge to try to get literal U+0027 in output when required,
+     * bypassing groff's Unicode transform; pinched from pod2man */
+    fprintf(fp, ".ie \\n(.g .ds Aq \\(aq\n"
+	        ".el       .ds Aq '\n");
+
     /* .TH name-of-program manual-section */
     fprintf(fp, ".TH");
     if (conf.th && *conf.th) {
@@ -418,7 +428,8 @@ void man_backend(paragraph *sourceform, keywordlist *keywords,
     /*
      * Tidy up.
      */
-    fclose(fp);
+    if (fp != stdout)
+	fclose(fp);
     man_conf_cleanup(conf);
 }
 
@@ -475,7 +486,8 @@ static int man_convert(wchar_t const *s, int maxlen,
 		 */
 		rdaddc(&out, '\\');
 		rdaddc(&out, '&');
-	    } else if (*q == '`' || *q == ' ') {
+	    }
+	    if (*q == '`' || *q == ' ') {
 		/* Quote backticks and nonbreakable spaces always. */
 		rdaddc(&out, '\\');
 	    } else if (*q == '\\') {
@@ -483,8 +495,26 @@ static int man_convert(wchar_t const *s, int maxlen,
 		rdaddsc(&out, "\\e");
 		continue;
 	    } else if (*q == '-') {
-		/* Turn nonbreakable hyphens into \(hy. */
-		rdaddsc(&out, "\\(hy");
+		if (quote_props & QUOTE_LITERAL) {
+		    /*
+		     * Try to preserve literal U+002D.
+		     * This is quite awkward. Debian hacks groff so that
+		     * \- and - both produce it; elsewhere it's not necessarily
+		     * possible to get it.
+		     * Apparently \- is the preferred compromise despite
+		     * having minus-sign semantics, as it is non-breaking.
+		     * (pod2man uses it, anyway.)
+		     */
+		    rdaddc(&out, '\\');
+		} else {
+		    /* Turn nonbreakable hyphens into \(hy. */
+		    rdaddsc(&out, "\\(hy");
+		    continue;
+		}
+	    } else if (*q == '\'' && (quote_props & QUOTE_LITERAL)) {
+		/* Try to preserve literal U+0027 (using string defined
+		 * in preamble) */
+		rdaddsc(&out, "\\*(Aq"); /* "apostrophe quote" */
 		continue;
 	    } else if (*q == '"' && (quote_props & QUOTE_QUOTES)) {
 		/*
@@ -560,14 +590,17 @@ static int man_rdaddwc(rdstringc *rs, word *text, word *end,
 
       case word_Normal:
       case word_Emph:
+      case word_Strong:
       case word_Code:
       case word_WeakCode:
       case word_WhiteSpace:
       case word_EmphSpace:
+      case word_StrongSpace:
       case word_CodeSpace:
       case word_WkCodeSpace:
       case word_Quote:
       case word_EmphQuote:
+      case word_StrongQuote:
       case word_CodeQuote:
       case word_WkCodeQuote:
 	assert(text->type != word_CodeQuote &&
@@ -577,6 +610,10 @@ static int man_rdaddwc(rdstringc *rs, word *text, word *end,
 	    (attraux(text->aux) == attr_First ||
 	     attraux(text->aux) == attr_Only)) {
 	    quote_props = man_rdaddctrl(rs, "\\fI", quote_props, conf, state);
+	} else if (towordstyle(text->type) == word_Strong &&
+	    (attraux(text->aux) == attr_First ||
+	     attraux(text->aux) == attr_Only)) {
+	    quote_props = man_rdaddctrl(rs, "\\fB", quote_props, conf, state);
 	} else if ((towordstyle(text->type) == word_Code ||
 		    towordstyle(text->type) == word_WeakCode) &&
 		   (attraux(text->aux) == attr_First ||
@@ -584,6 +621,10 @@ static int man_rdaddwc(rdstringc *rs, word *text, word *end,
 	    quote_props = man_rdaddctrl(rs, "\\fB", quote_props, conf, state);
 	}
 
+	if (towordstyle(text->type) == word_Code ||
+	    towordstyle(text->type) == word_WeakCode)
+	    quote_props |= QUOTE_LITERAL;
+
 	if (removeattr(text->type) == word_Normal) {
 	    charset_state s2 = *state;
 	    int len = ustrlen(text->text), hyphen = FALSE;
@@ -654,7 +695,7 @@ static void man_codepara(FILE *fp, word *text, int charset) {
     for (; text; text = text->next) if (text->type == word_WeakCode) {
 	char *c;
 	wchar_t *t, *e;
-	int quote_props = QUOTE_INITCTRL;
+	int quote_props = QUOTE_INITCTRL | QUOTE_LITERAL;
 
 	t = text->text;
 	if (text->next && text->next->type == word_Emph) {
