@@ -14,9 +14,6 @@ paragraph *pdf_config_filename(char *filename)
     return cmdline_cfg_simple("pdf-filename", filename, NULL);
 }
 
-typedef struct object_Tag object;
-typedef struct objlist_Tag objlist;
-
 struct object_Tag {
     objlist *list;
     object *next;
@@ -31,10 +28,6 @@ struct objlist_Tag {
     object *head, *tail;
 };
 
-static object *new_object(objlist *list);
-static void objtext(object *o, char const *text);
-static void objstream(object *o, char const *text);
-static void objstream_len(object *o, char const *text, size_t len);
 static void pdf_string(void (*add)(object *, char const *),
 		       object *, char const *);
 static void pdf_string_len(void (*add)(object *, char const *),
@@ -57,7 +50,6 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
     int font_index;
     font_encoding *fe;
     page_data *page;
-    int pageno;
     FILE *fp;
     char *filename;
     paragraph *p;
@@ -137,8 +129,11 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
     objtext(resources, "<<\n/ProcSet [/PDF/Text]\n/Font <<\n");
     for (fe = doc->fonts->head; fe; fe = fe->next) {
 	char fname[40];
+	char buf[80];
 	int i, prev;
-	object *font;
+	object *font, *fontdesc;
+	int flags;
+	font_info const *fi = fe->font->info;
 
 	sprintf(fname, "f%d", font_index++);
 	fe->name = dupstr(fname);
@@ -151,26 +146,13 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
 	objref(resources, font);
 	objtext(resources, "\n");
 
-	objtext(font, "<<\n/Type /Font\n/Subtype /Type1\n/Name /");
-	objtext(font, fe->name);
-	objtext(font, "\n/BaseFont /");
-	objtext(font, fe->font->info->name);
-	objtext(font, "\n/Encoding <<\n/Type /Encoding\n/Differences [");
 
-	for (i = 0; i < 256; i++) {
-	    char buf[20];
-	    if (fe->vector[i] == NOGLYPH)
-		continue;
-	    if (i != prev + 1) {
-		sprintf(buf, "\n%d", i);
-		objtext(font, buf);
-	    }
-	    objtext(font, i % 8 ? "/" : "\n/");
-	    objtext(font, glyph_extern(fe->vector[i]));
-	    prev = i;
-	}
-
-	objtext(font, "\n]\n>>\n");
+	/*
+	 * Construct those parts of the font descriptor that don't dependd
+	 * on the file format.
+	 */
+	if (!is_std_font(fe->font->info->name)) {
+	    fontdesc = new_object(&olist);
 
 #define FF_FIXEDPITCH	0x00000001
 #define FF_SERIF	0x00000002
@@ -182,36 +164,6 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
 #define FF_SMALLCAP	0x00020000
 #define FF_FORCEBOLD	0x00040000
 
-	if (!is_std_font(fe->font->info->name)){
-	    object *widths = new_object(&olist);
-	    object *fontdesc = new_object(&olist);
-	    int firstchar = -1, lastchar = -1;
-	    char buf[80];
-	    font_info const *fi = fe->font->info;
-	    int flags;
-	    for (i = 0; i < 256; i++)
-		if (fe->vector[i] != NOGLYPH) {
-		    if (firstchar < 0) firstchar = i;
-		    lastchar = i;
-		}
-	    sprintf(buf, "/FirstChar %d\n/LastChar %d\n/Widths ",
-		    firstchar, lastchar);
-	    objtext(font, buf);
-	    objref(font, widths);
-	    objtext(font, "\n");
-	    objtext(widths, "[\n");
-	    for (i = firstchar; i <= lastchar; i++) {
-		double width;
-		if (fe->vector[i] == NOGLYPH)
-		    width = 0.0;
-		else
-		    width = find_width(fe->font, fe->vector[i]);
-		sprintf(buf, "%g\n", 1000.0 * width / FUNITS_PER_PT);
-		objtext(widths, buf);
-	    }
-	    objtext(widths, "]\n");
-	    objtext(font, "/FontDescriptor ");
-	    objref(font, fontdesc);
 	    objtext(fontdesc, "<<\n/Type /FontDescriptor\n/Name /");
 	    objtext(fontdesc, fi->name);
 	    flags = 0;
@@ -236,7 +188,181 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
 	    objtext(fontdesc, buf);
 	    sprintf(buf, "/StemV %g\n", fi->stemv);
 	    objtext(fontdesc, buf);
-	    if (fi->fontfile) {
+	}
+
+	objtext(font, "<<\n/Type /Font\n/BaseFont /");
+	objtext(font, fe->font->info->name);
+	if (fe->font->info->filetype == TRUETYPE) {
+	    object *cidfont = new_object(&olist);
+	    object *cmap = new_object(&olist);
+	    unsigned short ranges[256];
+	    unsigned startidx, nranges, nchars;
+	    int start;
+
+	    objtext(font, "/Subtype/Type0\n/Encoding ");
+	    objtext(cmap, "<</Type/CMap\n/CMapName/");
+	    objtext(cmap, fe->name);
+	    objtext(cmap, "\n/CIDSystemInfo<</Registry(Adobe)"
+		    "/Ordering(Identity)/Supplement 0>>\n");
+	    objstream(cmap, "%!PS-Adobe-3.0 Resource-CMap\n"
+		      "%%DocumentNeededResources: procset CIDInit\n"
+		      "%%IncludeResource: procset CIDInit\n"
+		      "%%BeginResource: CMap ");
+	    objstream(cmap, fe->name);
+	    objstream(cmap, "\n%%Title (");
+	    objstream(cmap, fe->name);
+	    objstream(cmap, " Adobe Identity 0)\n%%Version: 1\n%%EndComments\n");
+	    objstream(cmap, "/CIDInit/ProcSet findresource begin\n");
+	    objstream(cmap, "12 dict begin begincmap\n");
+	    objstream(cmap, "/CIDSystemInfo 3 dict dup begin\n"
+		      "/Registry(Adobe)def/Ordering(Identity)def"
+		      "/Supplement 0 def end def\n");
+	    objstream(cmap, "/CMapName/");
+	    objstream(cmap, fe->name);
+	    objstream(cmap, " def/CMapType 0 def/WMode 0 def\n");
+	    objstream(cmap, "1 begincodespacerange<00><FF>"
+		      "endcodespacerange\n");
+	    start = -1; nranges = nchars = 0;
+	    for (i = 0; i < 256; i++) {
+		unsigned idx;
+
+		ranges[i] = 0;
+		if (fe->vector[i] == NOGLYPH)
+		    continue;
+		idx = sfnt_glyphtoindex(fe->font->info->fontfile,
+					fe->vector[i]);
+		if (start >= 0 && idx - startidx == (unsigned)(i - start)) {
+		    if (ranges[start] == 1) {
+			nranges++; nchars--;
+		    }
+		    ranges[start] = i - start + 1;
+		} else {
+		    ranges[i] = 1;
+		    start = i;
+		    startidx = idx;
+		    nchars++;
+		}
+	    }
+	    i = 0;
+	    while (nranges) {
+		unsigned blk = nranges > 100 ? 100 : nranges;
+		nranges -= blk;
+		sprintf(buf, "%u ", blk);
+		objstream(cmap, buf);
+		objstream(cmap, "begincidrange\n");
+		while (blk) {
+		    if (ranges[i] > 1) {
+			sprintf(buf, "<%02X>", i);
+			objstream(cmap, buf);
+			sprintf(buf, "<%02X>", i + ranges[i] - 1);
+			objstream(cmap, buf);
+			sprintf(buf, "%hu\n",
+				(unsigned short)
+				sfnt_glyphtoindex(fe->font->info->fontfile,
+						  fe->vector[i]));
+			objstream(cmap, buf);
+			blk--;
+		    }
+		    i++;
+		}
+		objstream(cmap, "endcidrange\n");
+	    }
+	    i = 0;
+	    while (nchars) {
+		unsigned blk = nchars > 100 ? 100 : nchars;
+		nchars -= blk;
+		sprintf(buf, "%u ", blk);
+		objstream(cmap, buf);
+		objstream(cmap, "begincidchar\n");
+		while (blk) {
+		    if (ranges[i] == 1) {
+			sprintf(buf, "<%02X>", i);
+			objstream(cmap, buf);
+			sprintf(buf, "%hu\n",
+				(unsigned short)
+				sfnt_glyphtoindex(fe->font->info->fontfile,
+						  fe->vector[i]));
+			objstream(cmap, buf);
+			blk--;
+		    }
+		    i++;
+		}
+		objstream(cmap, "endcidchar\n");
+	    }
+	    objstream(cmap, "endcmap CMapName currentdict /CMap "
+		      "defineresource pop end end\n%%EndResource\n%%EOF\n");
+
+	    objref(font, cmap);
+	    objtext(font, "\n/DescendantFonts[");
+	    objref(font, cidfont);
+	    objtext(font, "]\n");
+	    objtext(cidfont, "<<\n/Type/Font\n/Subtype/CIDFontType2\n"
+		    "/BaseFont/");
+	    objtext(cidfont, fe->font->info->name);
+	    objtext(cidfont, "\n/CIDSystemInfo<</Registry(Adobe)"
+		    "/Ordering(Identity)/Supplement 0>>\n");
+	    objtext(cidfont, "/FontDescriptor ");
+	    objref(cidfont, fontdesc);
+	    objtext(cidfont, "\n/W[0[");
+	    for (i = 0; i < (int)sfnt_nglyphs(fe->font->info->fontfile); i++) {
+		char buf[20];
+		double width;
+		width = find_width(fe->font,
+			       sfnt_indextoglyph(fe->font->info->fontfile, i));
+		sprintf(buf, "%g ", 1000.0 * width / FUNITS_PER_PT);
+		objtext(cidfont, buf);
+	    }
+	    objtext(cidfont, "]]>>\n");
+	} else {
+	    objtext(font, "/Subtype /Type1\n");
+	    objtext(font, "\n/Encoding <<\n/Type /Encoding\n/Differences [");
+
+	    for (i = 0; i < 256; i++) {
+		char buf[20];
+		if (fe->vector[i] == NOGLYPH)
+		    continue;
+		if (i != prev + 1) {
+		    sprintf(buf, "\n%d", i);
+		    objtext(font, buf);
+		}
+		objtext(font, i % 8 ? "/" : "\n/");
+		objtext(font, glyph_extern(fe->vector[i]));
+		prev = i;
+	    }
+
+	    objtext(font, "\n]\n>>\n");
+	    if (!is_std_font(fe->font->info->name)){
+		object *widths = new_object(&olist);
+		int firstchar = -1, lastchar = -1;
+		for (i = 0; i < 256; i++)
+		    if (fe->vector[i] != NOGLYPH) {
+			if (firstchar < 0) firstchar = i;
+			lastchar = i;
+		    }
+		sprintf(buf, "/FirstChar %d\n/LastChar %d\n/Widths ",
+			firstchar, lastchar);
+		objtext(font, buf);
+		objref(font, widths);
+		objtext(font, "\n");
+		objtext(widths, "[\n");
+		for (i = firstchar; i <= lastchar; i++) {
+		    double width;
+		    if (fe->vector[i] == NOGLYPH)
+			width = 0.0;
+		    else
+			width = find_width(fe->font, fe->vector[i]);
+		    sprintf(buf, "%g\n", 1000.0 * width / FUNITS_PER_PT);
+		    objtext(widths, buf);
+		}
+		objtext(widths, "]\n");
+		objtext(font, "/FontDescriptor ");
+		objref(font, fontdesc);
+	    }
+
+	}
+
+	if (!is_std_font(fe->font->info->name)) {
+	    if (fi->fontfile && fi->filetype == TYPE1) {
 		object *fontfile = new_object(&olist);
 		size_t len;
 		char *ffbuf;
@@ -254,6 +380,17 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
 		objtext(fontfile, "/Length3 0\n");
 		objtext(fontdesc, "/FontFile ");
 		objref(fontdesc, fontfile);
+	    } else if (fi->fontfile && fi->filetype == TRUETYPE) {
+		object *fontfile = new_object(&olist);
+		size_t len;
+		char *ffbuf;
+
+		sfnt_data((font_info *)fi, &ffbuf, &len);
+		objstream_len(fontfile, ffbuf, len);
+		sprintf(buf, "<<\n/Length1 %lu\n", (unsigned long)len);
+		objtext(fontfile, buf);
+		objtext(fontdesc, "/FontFile2 ");
+		objref(fontdesc, fontfile);
 	    }
 	    objtext(fontdesc, "\n>>\n");
 	}
@@ -291,7 +428,6 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
     /*
      * Create and render the individual pages.
      */
-    pageno = 0;
     for (page = doc->pages; page; page = page->next) {
 	object *opage, *cstr;
 	rect *r;
@@ -494,7 +630,7 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
 	    zbuf = snewn(zlen, char);
 	    memcpy(zbuf, o->stream.text, zlen);
 	    sprintf(text, "/Length %d\n>>\n", zlen);
-#else	    
+#else
 	    zcontext = deflate_compress_new(DEFLATE_TYPE_ZLIB);
 	    deflate_compress_data(zcontext, o->stream.text, o->stream.pos,
 				  DEFLATE_END_OF_DATA, &zbuf, &zlen);
@@ -529,9 +665,12 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
      * Write out the PDF file.
      */
 
-    fp = fopen(filename, "wb");
+    if (!strcmp(filename, "-"))
+	fp = stdout;
+    else
+	fp = fopen(filename, "wb");
     if (!fp) {
-	error(err_cantopenw, filename);
+	err_cantopenw(filename);
 	return;
     }
 
@@ -576,12 +715,13 @@ void pdf_backend(paragraph *sourceform, keywordlist *keywords,
 	    olist.tail->number + 1, cat->number, info->number);
     fprintf(fp, "startxref\n%d\n%%%%EOF\n", fileoff);
 
-    fclose(fp);
+    if (fp != stdout)
+	fclose(fp);
 
     sfree(filename);
 }
 
-static object *new_object(objlist *list)
+object *new_object(objlist *list)
 {
     object *obj = snew(object);
 
@@ -607,17 +747,17 @@ static object *new_object(objlist *list)
     return obj;
 }
 
-static void objtext(object *o, char const *text)
+void objtext(object *o, char const *text)
 {
     rdaddsc(&o->main, text);
 }
 
-static void objstream_len(object *o, char const *text, size_t len)
+void objstream_len(object *o, char const *text, size_t len)
 {
     rdaddsn(&o->stream, text, len);
 }
 
-static void objstream(object *o, char const *text)
+void objstream(object *o, char const *text)
 {
     rdaddsc(&o->stream, text);
 }
