@@ -5,13 +5,19 @@
 #include <stdio.h>
 #include "halibut.h"
 
-static char *helptext[] = {
+static const char *const helptext[] = {
     "usage:   halibut [options] files",
     "options: --text[=filename]     generate plain text output",
     "         --html[=filename]     generate XHTML output",
     "         --winhelp[=filename]  generate Windows Help output",
     "         --man[=filename]      generate man page output",
+    "         --info[=filename]     generate GNU info output",
+    "         --ps[=filename]       generate PostScript output",
+    "         --pdf[=filename]      generate PDF output",
     "         -Cfoo:bar:baz         append \\cfg{foo}{bar}{baz} to input",
+    "         --input-charset=cs    change default input file charset",
+    "         --list-charsets       display supported character set names",
+    "         --list-fonts          display supported font names",
     "         --precise             report column numbers in error messages",
     "         --help                display this text",
     "         --version             display version number",
@@ -19,19 +25,19 @@ static char *helptext[] = {
     NULL
 };
 
-static char *usagetext[] = {
-    "usage: halibut [--format[=filename]] [-Cconfig...] file.but [file.but...]",
+static const char *const usagetext[] = {
+    "usage: halibut [--format[=filename]] [options] file.but [file.but...]",
     NULL
 };
 
 void help(void) {
-    char **p;
+    const char *const *p;
     for (p = helptext; *p; p++)
 	puts(*p);
 }
 
 void usage(void) {
-    char **p;
+    const char *const *p;
     for (p = usagetext; *p; p++)
 	puts(*p);
 }
@@ -39,3 +45,13 @@ void usage(void) {
 void showversion(void) {
     printf("Halibut, %s\n", version);
 }
+
+void listcharsets(void) {
+    int i = 0, c;
+    do {
+	c = charset_localenc_nth(i);
+	if (c == CS_NONE) break;
+	printf("%s\n", charset_to_localenc(c));
+	i++;
+    } while (1);
+}
