@@ -49,7 +49,7 @@ static void macrodef(tree234 *macros, wchar_t *name, wchar_t *text,
     m->name = name;
     m->text = text;
     if (add234(macros, m) != m) {
-	error(err_macroexists, &fpos, name);
+	err_macroexists(&fpos, name);
 	sfree(name);
 	sfree(text);
     }
@@ -119,7 +119,8 @@ static int get(input *in, filepos *pos, rdstringc *rsc) {
 	    int c = getc(in->currfp);
 
 	    if (c == EOF) {
-		fclose(in->currfp);
+		if (in->wantclose)
+		    fclose(in->currfp);
 		in->currfp = NULL;
 		return EOF;
 	    }
@@ -238,6 +239,7 @@ enum {
     c_q,			       /* quote marks */
     c_quote,			       /* block-quoted paragraphs */
     c_rule,			       /* horizontal rule */
+    c_s,			       /* strong */
     c_title,			       /* document title */
     c_u,			       /* aux field is char code */
     c_versionid			       /* document RCS id */
@@ -309,6 +311,7 @@ static void match_kw(token *tok) {
 	{"q", c_q},		       /* quote marks */
 	{"quote", c_quote},	       /* block-quoted paragraphs */
 	{"rule", c_rule},	       /* horizontal rule */
+	{"s", c_s},		       /* strong */
 	{"title", c_title},	       /* document title */
 	{"versionid", c_versionid},    /* document RCS id */
 	{"{", c__escaped},	       /* escaped lbrace (\{) */
@@ -691,8 +694,11 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		} else if (t.type == tok_cmd && t.cmd == c_e &&
 			   wtype == word_WeakCode) {
 		    wtype = word_Emph;
+		} else if (t.type == tok_cmd && t.cmd == c_s &&
+			   wtype == word_WeakCode) {
+		    wtype = word_Strong;
 		} else {
-		    error(err_brokencodepara, &t.pos);
+		    err_brokencodepara(&t.pos);
 		    prev_para_type = par.type;
 		    addpara(par, ret);
 		    while (t.type != tok_eop)   /* error recovery: */
@@ -721,7 +727,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 	     */
 	    dtor(t), t = get_token(in);
 	    if (t.type != tok_lbrace) {
-		error(err_explbr, &t.pos);
+		err_explbr(&t.pos);
 		continue;
 	    }
 
@@ -764,7 +770,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		     * don't give a cascade error.
 		     */
 		    sitem->type = -1;
-		    error(err_misplacedlcont, &t.pos);
+		    err_misplacedlcont(&t.pos);
 		}
 	    } else {
 		/*
@@ -789,7 +795,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 	} else if (t.type == tok_rbrace) {
 	    struct crossparaitem *sitem = stk_pop(crossparastk);
 	    if (!sitem)
-		error(err_unexbrace, &t.pos);
+		err_unexbrace(&t.pos);
 	    else {
 		switch (sitem->type) {
 		  case c_lcont:
@@ -831,12 +837,14 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		needkw = -1;
 		break;
 	      case c__invalid:
-		error(err_badparatype, t.text, &t.pos);
+		err_badparatype(t.text, &t.pos);
 		needkw = 4;
 		break;
 	      case c__comment:
-		if (isbrace(in))
+		if (isbrace(in)) {
+		    needkw = -1;
 		    break;	       /* `\#{': isn't a comment para */
+		}
 		do {
 		    dtor(t), t = get_token(in);
 		} while (t.type != tok_eop && t.type != tok_eof);
@@ -888,8 +896,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		par.type == para_UnnumberedChapter) {
 		struct crossparaitem *sitem = stk_top(crossparastk);
 		if (sitem && (sitem->seen_lcont || sitem->seen_quote)) {
-		    error(err_sectmarkerinblock,
-			  &t.pos,
+		    err_sectmarkerinblock(			  &t.pos,
 			  (sitem->seen_lcont ? "lcont" : "quote"));
 		}
 	    }
@@ -942,7 +949,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			}
 		    }
 		    if (t.type != tok_rbrace) {
-			error(err_kwunclosed, &t.pos);
+			err_kwunclosed(&t.pos);
 			continue;
 		    }
 		    rdadd(&rs, 0);     /* add string terminator */
@@ -955,11 +962,11 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 
 		/* See whether we have the right number of keywords. */
 		if ((needkw & 48) && nkeys > 0)
-		    error(err_kwillegal, &fp);
+		    err_kwillegal(&fp);
 		if ((needkw & 11) && nkeys == 0)
-		    error(err_kwexpected, &fp);
+		    err_kwexpected(&fp);
 		if ((needkw & 5) && nkeys > 1)
-		    error(err_kwtoomany, &fp);
+		    err_kwtoomany(&fp);
 
 		if (is_macro) {
 		    /*
@@ -975,7 +982,8 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			    rdadd(&macrotext, L'\n');
 			rdadds(&macrotext, t.text);
 			dtor(t), t = get_token(in);
-			if (t.type == tok_eop) break;
+			if (t.type == tok_eop || t.type == tok_eof)
+                            break;
 		    }
 		    macrodef(macros, rs.text, macrotext.text, fp);
 		    continue;	       /* next paragraph */
@@ -992,7 +1000,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		    if (t.type != tok_eop && t.type != tok_eof &&
 			(start_cmd == c__invalid ||
 			 t.type != tok_cmd || t.cmd != start_cmd)) {
-			error(err_bodyillegal, &t.pos);
+			err_bodyillegal(&t.pos);
 			/* Error recovery: eat the rest of the paragraph */
 			while (t.type != tok_eop && t.type != tok_eof &&
 			       (start_cmd == c__invalid ||
@@ -1109,7 +1117,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		}
 		break;
 	      case tok_lbrace:
-		error(err_unexbrace, &t.pos);
+		err_unexbrace(&t.pos);
 		/* Error recovery: push nop */
 		sitem = snew(struct stack_item);
 		sitem->type = stack_nop;
@@ -1195,7 +1203,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		     */
 		    dtor(t), t = get_token(in);
 		    if (t.type != tok_lbrace) {
-			error(err_explbr, &t.pos);
+			err_explbr(&t.pos);
 		    } else {
 			int braces = 1;
 			while (braces > 0) {
@@ -1205,7 +1213,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			    else if (t.type == tok_rbrace)
 				braces--;
 			    else if (t.type == tok_eof) {
-				error(err_commenteof, &t.pos);
+				err_commenteof(&t.pos);
 				break;
 			    }
 			}
@@ -1224,7 +1232,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
                     type = t.cmd;
 		    dtor(t), t = get_token(in);
 		    if (t.type != tok_lbrace) {
-			error(err_explbr, &t.pos);
+			err_explbr(&t.pos);
 		    } else {
 			/*
 			 * Enforce that \q may not be used anywhere
@@ -1253,7 +1261,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			    }
 			    stype = stack_quote;
 			} else {
-			    error(err_codequote, &t.pos);
+			    err_codequote(&t.pos);
 			    stype = stack_nop;
 			}
 			sitem = snew(struct stack_item);
@@ -1261,7 +1269,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			sitem->type = stype;
                         if (type == c_cq) {
                             if (style != word_Normal) {
-                                error(err_nestedstyles, &t.pos);
+                                err_nestedstyles(&t.pos);
                             } else {
                                 style = word_WeakCode;
                                 spcstyle = tospacestyle(style);
@@ -1299,7 +1307,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			    wdtext = ustrftime(NULL, broken);
 			    wd.type = style;
 			} else {
-			    error(err_explbr, &t.pos);
+			    err_explbr(&t.pos);
 			    wdtext = NULL;
 			}
 		    } else {
@@ -1321,7 +1329,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			}
 			sfree(rs.text);
 			if (t.type != tok_rbrace) {
-			    error(err_kwexprbr, &t.pos);
+			    err_kwexprbr(&t.pos);
 			}
 		    }
 		    wd.alt = NULL;
@@ -1351,7 +1359,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			if (t.type == tok_cmd &&
 			    (t.cmd == c_i || t.cmd == c_ii)) {
 			    if (indexing) {
-				error(err_nestedindex, &t.pos);
+				err_nestedindex(&t.pos);
 			    } else {
 				/* Add an index-reference word with no
 				 * text as yet */
@@ -1377,15 +1385,17 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			    dtor(t), t = get_token(in);
 			}
 			/*
-			 * Special cases: \W{}\c, \W{}\e, \W{}\cw
+			 * Special cases: \W{}\c, \W{}\e, \W{}\s, \W{}\cw
 			 */
 			if (t.type == tok_cmd &&
-			    (t.cmd == c_e || t.cmd == c_c || t.cmd == c_cw)) {
+			    (t.cmd == c_e || t.cmd == c_s ||
+                             t.cmd == c_c || t.cmd == c_cw)) {
 			    if (style != word_Normal)
-				error(err_nestedstyles, &t.pos);
+				err_nestedstyles(&t.pos);
 			    else {
 				style = (t.cmd == c_c ? word_Code :
 					 t.cmd == c_cw ? word_WeakCode :
+					 t.cmd == c_s ? word_Strong :
 					 word_Emph);
 				spcstyle = tospacestyle(style);
 				sitem->type |= stack_style;
@@ -1393,7 +1403,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 			    dtor(t), t = get_token(in);
 			}
 			if (t.type != tok_lbrace) {
-			    error(err_explbr, &t.pos);
+			    err_explbr(&t.pos);
 			    sfree(sitem);
 			} else {
 			    stk_push(parsestk, sitem);
@@ -1403,9 +1413,10 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		  case c_c:
 		  case c_cw:
 		  case c_e:
+		  case c_s:
 		    type = t.cmd;
 		    if (style != word_Normal) {
-			error(err_nestedstyles, &t.pos);
+			err_nestedstyles(&t.pos);
 			/* Error recovery: eat lbrace, push nop. */
 			dtor(t), t = get_token(in);
 			sitem = snew(struct stack_item);
@@ -1415,10 +1426,11 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		    }
 		    dtor(t), t = get_token(in);
 		    if (t.type != tok_lbrace) {
-			error(err_explbr, &t.pos);
+			err_explbr(&t.pos);
 		    } else {
 			style = (type == c_c ? word_Code :
 				 type == c_cw ? word_WeakCode :
+				 type == c_s ? word_Strong :
 				 word_Emph);
 			spcstyle = tospacestyle(style);
 			sitem = snew(struct stack_item);
@@ -1432,7 +1444,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		  case c_I:
 		    type = t.cmd;
 		    if (indexing) {
-			error(err_nestedindex, &t.pos);
+			err_nestedindex(&t.pos);
 			/* Error recovery: eat lbrace, push nop. */
 			dtor(t), t = get_token(in);
 			sitem = snew(struct stack_item);
@@ -1445,16 +1457,18 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		    sitem->type = stack_idx;
 		    dtor(t), t = get_token(in);
 		    /*
-		     * Special cases: \i\c, \i\e, \i\cw
+		     * Special cases: \i\c, \i\e, \i\s, \i\cw
 		     */
 		    wd.fpos = t.pos;
 		    if (t.type == tok_cmd &&
-			(t.cmd == c_e || t.cmd == c_c || t.cmd == c_cw)) {
+			(t.cmd == c_e || t.cmd == c_s ||
+                         t.cmd == c_c || t.cmd == c_cw)) {
 			if (style != word_Normal)
-			    error(err_nestedstyles, &t.pos);
+			    err_nestedstyles(&t.pos);
 			else {
 			    style = (t.cmd == c_c ? word_Code :
 				     t.cmd == c_cw ? word_WeakCode :
+				     t.cmd == c_s ? word_Strong :
 				     word_Emph);
 			    spcstyle = tospacestyle(style);
 			    sitem->type |= stack_style;
@@ -1463,7 +1477,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		    }
 		    if (t.type != tok_lbrace) {
 			sfree(sitem);
-			error(err_explbr, &t.pos);
+			err_explbr(&t.pos);
 		    } else {
 			/* Add an index-reference word with no text as yet */
 			wd.type = word_IndexRef;
@@ -1526,7 +1540,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 		    break;
 		  default:
 		    if (!macrolookup(macros, in, t.text, &t.pos))
-			error(err_badmidcmd, t.text, &t.pos);
+			err_badmidcmd(t.text, &t.pos);
 		    break;
 		}
 	    }
@@ -1539,7 +1553,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 	if (stk_top(parsestk)) {
 	    while ((sitem = stk_pop(parsestk)))
 		sfree(sitem);
-	    error(err_missingrbrace, &t.pos);
+	    err_missingrbrace(&t.pos);
 	}
 	stk_free(parsestk);
 	prev_para_type = par.type;
@@ -1561,7 +1575,7 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
     if (stk_top(crossparastk)) {
 	void *p;
 
-	error(err_missingrbrace2, &t.pos);
+	err_missingrbrace2(&t.pos);
 	while ((p = stk_pop(crossparastk)))
 	    sfree(p);
     }
@@ -1578,12 +1592,15 @@ static void read_file(paragraph ***ret, input *in, indexdata *idx,
 struct {
     char const *magic;
     size_t nmagic;
+    int binary;
     void (*reader)(input *);
 } magics[] = {
-    { "%!FontType1-",     12, &read_pfa_file },
-    { "%!PS-AdobeFont-",  15, &read_pfa_file },
-    { "\x80\x01",          2, &read_pfb_file },
-    { "StartFontMetrics", 16, &read_afm_file },
+    { "%!FontType1-",     12, FALSE, &read_pfa_file },
+    { "%!PS-AdobeFont-",  15, FALSE, &read_pfa_file },
+    { "\x80\x01",          2, TRUE,  &read_pfb_file },
+    { "StartFontMetrics", 16, FALSE, &read_afm_file },
+    { "\x00\x01\x00\x00",  4, TRUE,  &read_sfnt_file },
+    { "true",		   4, TRUE,  &read_sfnt_file },
 };
 
 paragraph *read_input(input *in, indexdata *idx) {
@@ -1592,33 +1609,64 @@ paragraph *read_input(input *in, indexdata *idx) {
     tree234 *macros;
     char mag[16];
     size_t len, i;
+    int binary;
     void (*reader)(input *);
 
     macros = newtree234(macrocmp);
 
     while (in->currindex < in->nfiles) {
-	in->currfp = fopen(in->filenames[in->currindex], "r");
-	if (in->currfp) {
-	    setpos(in, in->filenames[in->currindex]);
-	    in->charset = in->defcharset;
-	    in->csstate = charset_init_state;
-	    in->wcpos = in->nwc = 0;
-	    in->pushback_chars = NULL;
+	setpos(in, in->filenames[in->currindex]);
+	in->charset = in->defcharset;
+	in->csstate = charset_init_state;
+	in->wcpos = in->nwc = 0;
+	in->pushback_chars = NULL;
+
+	if (!in->filenames[in->currindex]) {
+	    in->currfp = stdin;
+	    in->wantclose = FALSE;     /* don't fclose stdin */
+	    /*
+	     * When reading standard input, we always expect to see
+	     * an actual Halibut file and not any of the unusual
+	     * input types like fonts.
+	     */
 	    reader = NULL;
-	    len = fread(mag, 1, sizeof(mag), in->currfp);
-	    for (i = 0; i < lenof(magics); i++) {
-		if (len >= magics[i].nmagic &&
-		    memcmp(mag, magics[i].magic, magics[i].nmagic) == 0) {
-		    reader = magics[i].reader;
-		    break;
+	} else {
+	    /*
+	     * Open the file in binary mode to look for magic
+	     * numbers. We'll switch to text mode if we find we're
+	     * looking at a text file type.
+	     */
+	    in->currfp = fopen(in->filenames[in->currindex], "rb");
+	    binary = FALSE; /* default to Halibut source, which is text */
+	    if (in->currfp) {
+		in->wantclose = TRUE;
+		reader = NULL;
+		len = fread(mag, 1, sizeof(mag), in->currfp);
+		for (i = 0; i < lenof(magics); i++) {
+		    if (len >= magics[i].nmagic &&
+			memcmp(mag, magics[i].magic, magics[i].nmagic) == 0) {
+			reader = magics[i].reader;
+			binary = magics[i].binary;
+			break;
+		    }
 		}
+		rewind(in->currfp);
+	    }
+	    if (!binary) {
+                if (in->currfp)
+                    fclose(in->currfp);
+		in->currfp = fopen(in->filenames[in->currindex], "r");
 	    }
-	    rewind(in->currfp);
-	    if (reader == NULL)
-		read_file(&hptr, in, idx, macros);
-	    else
-		(*reader)(in);
 	}
+	if (in->currfp) {
+	    if (reader == NULL) {
+		read_file(&hptr, in, idx, macros);
+	    } else {
+		(*reader)(in);
+	    }
+	} else {
+            err_cantopen(in->filenames[in->currindex]);
+        }
 	in->currindex++;
     }
 
